# required libraries 
library(LMest) 
library(numDeriv)
library(corpcor)

## main function: lmFirth
## 

lmFirth = function(S, X, lev, k, init = 1, mx = 10, inits = NULL, make.pos.def=TRUE) {
       n=dim(X)[1]
    if (k > 1) {
        TT = dim(X)[2]
        p = nc = dim(X)[3]
        ne = lev - 1
        nt = prod(lev)
        if (init == 0) {
            inits = est_lm_cov_manifest(S, X, lev, k, 1, mod = 0, maxit = 0)
        }
        if (init == 1) {
            inits = est_lm_cov_manifest(S, X, lev, k, 1, mod = 0, start = 0, maxit = mx)
        }
        if (init == 2) {
            inits = est_lm_cov_manifest(S, X, lev, k, 1, mod = 0, start = 1, maxit = mx)
        }
        if (is.array(S)) 
            S = matrix(S, n, TT)
        Y0 = S + 1
        S = array(0, c(nt, n, TT))
        for (i in 1:n) for (t in 1:TT) {
            ind = Y0[i, t]
            S[ind, i, t] = 1
        }
        XX = X
        X = array(0, c(ne, nc, n, TT))
        for (i in 1:n) for (t in 1:TT) {
            X[, , i, t] = rep(1, ne) %o% XX[i, t, ]
        }
        lap = log(inits$la[-1]/inits$la[1])
        pip = matrix(NA, k, k - 1)
        for (j in 1:k) {
            pip[j, ] = log(inits$PI[j, -j]/inits$PI[j, j])
        }
        th = c(inits$al, inits$be, lap, as.vector(pip), inits$mu)
        
        obj = function(th) {
            al = th[1:k]
            be = th[k + (1:p)]
            la = th[k + p + 1:(k - 1)]
            la = c(1, exp(la))/(1 + sum(exp(la)))
            pip = matrix(th[2 * k - 1 + p + 1:(k * (k - 1))], k, k - 1)
            PI = matrix(NA, k, k)
            for (j in 1:k) {
                sm = 1 + sum(exp(pip[j, ]))
                PI[j, -j] = exp(pip[j, ])/sm
                PI[j, j] = 1/sm
            }
            mu = th[2 * k - 1 + k * (k - 1) + p + 1:(lev - 1)]
            a = est_lm_cov_manifest2(S, X, lev, k, q = 1, mod = 0, maxit = 0, out_se = TRUE, 
                al = al, be = be, la = la, PI = PI, mu = mu, start = 2)
            J1 = a$J1
            lk = a$lk
            # firth pen link:
            ld = log(det(J1))
            if(any(apply(J1,1,sum)==0)) {
            w = which(apply(J1, 1, sum) == 0)
            J1 = J1[-w,-w]
            ld = log(det(J1))}
            if(make.pos.def) {
                if(!is.positive.definite(J1)) {
                    J1=make.positive.definite(J1)
                    ld = log(det(J1))}
            }
            if (is.nan(ld)) {
                ld = -1e9
            }
            if (is.nan(lk)) {
                return(-Inf)
            }
            return(lk + 0.5 * ld)
        }
        
        op = optim(th, function(x) -obj(x))
        
        th = op$par
        al = th[1:k]
        be = th[k + (1:p)]
        la = th[k + p + 1:(k - 1)]
        la = c(1, exp(la))/(1 + sum(exp(la)))
        pip = matrix(th[2 * k - 1 + p + 1:(k * (k - 1))], k, k - 1)
        PI = matrix(NA, k, k)
        for (j in 1:k) {
            sm = 1 + sum(exp(pip[j, ]))
            PI[j, -j] = exp(pip[j, ])/sm
            PI[j, j] = 1/sm
        }
        mu = th[2 * k - 1 + k * (k - 1) + p + 1:(lev - 1)]
        vals = est_lm_cov_manifest2(S, X, lev, k, 1, mod = 0, maxit = 0, out_se = TRUE, 
            al = al, be = be, la = la, PI = PI, mu = mu)
        al=al+mu[length(mu)]
        mu=mu-mu[length(mu)]
        return(list(al = al, be = be, la = la, PI = PI, mu = mu, lk = vals$lk, J = vals$J1, 
            score = vals$score))
    }
    if (k == 1) {
        lmFirth1(S, X, lev, k = 1, init, mx, inits)
    }
}

ticc=function(esta,make.pos.def=TRUE) {
J=esta$score
if(any(J==0)) {J=J[-which(J==0)]}
J=J%*%t(J)
JJ=esta$J
if(any(apply(abs(JJ),1,sum)==0)) {
w=which(apply(abs(JJ),1,sum)==0)
JJ=JJ[-w,-w]}
pen=Inf
if(make.pos.def) {
if(!is.positive.definite(J)) {
J=make.positive.definite(J)}
if(!is.positive.definite(JJ)) {
JJ=make.positive.definite(JJ)}}
JJ=try(solve(JJ))
if(!inherits(JJ,"try-error")) {
pen=sum(diag(J%*%JJ))*2}
c(-2*esta$lk+pen,esta$lk)}

lmFirth1 = function(S, X, lev, k = 1, init = 1, mx = 10, inits = NULL) {
    n = dim(X)[1]
    TT = dim(X)[2]
    p = nc = dim(X)[3]
    ne = lev - 1
    nt = prod(lev)
    if (init == 0) {
        inits = est_lm_cov_manifest(S, X, lev, k, 1, mod = 0, maxit = 0)
    }
    if (init == 1) {
        inits = est_lm_cov_manifest(S, X, lev, k, 1, mod = 0, maxit = mx, start = 1)
    }
    
    th = c(inits$be, inits$mu)
    
    obj = function(th) {
        be = th[1:p]
        mu = th[p + 1:(lev - 1)]
        a = est_lm_cov_manifest(S, X, lev, 1, 1, mod = 0, maxit = 1, out_se = TRUE, 
            la = 1, al = 0, PI = 1, be = be, mu = mu, start = 2)
        J1 = a$J1
        lk = a$lk
        # firth pen link:
        ld = log(det(J1))
        if (any(apply(J1, 1, sum) == 0 | is.nan(apply(J1, 1, sum)))) {
            w = which(apply(J1, 1, sum) == 0 | is.nan(apply(J1, 1, sum)))
            ld = log(det(J1[-w, -w]))
        }
        if (is.nan(ld)) {
            ld = -1e-9
        }
        if (is.nan(lk)) {
            return(-Inf)
        }
        return(lk + 0.5 * ld)
    }
    
    op = optim(th, function(x) -obj(x))
    
    th = op$par
    be = th[1:p]
    mu = th[p + 1:(lev - 1)]
    vals = est_lm_cov_manifest(S, X, lev, 1, 1, mod = 0, maxit = 0, out_se = TRUE, 
        be = be, mu = mu, la = 1, al = 0, PI = 1)
        mu=mu-mu[length(mu)]
    return(list(be = be, mu = mu, lk = vals$lk, J = vals$J1, score = vals$score, 
        PRED1 = vals$PRED1))
} 


## AUXILIARY FUNCTIONS ## 

# functions from package LMest 
blkdiag = LMest:::blkdiag 
expit = LMest:::expit 
rec1 = LMest:::rec1 
rec3 = LMest:::rec3 
lk_sta = LMest:::lk_sta  
lk_obs_manifest = LMest:::lk_obs_manifest 
lk_sta = LMest:::lk_sta  

# this rewrites function from package LMest so that it is quicker for
#  our purposes

est_lm_cov_manifest = function(S, X, lev, k, q, mod, tol = 10^-8, maxit = 1000, start = 0, 
    mu = NULL, al = NULL, be = NULL, si = NULL, rho = NULL, la = NULL, PI = NULL, 
    output = FALSE, out_se = FALSE) {
    nt = prod(lev)
    n = nrow(S)
    TT = ncol(S)
    if (is.array(S)) 
        S = matrix(S, n, TT)
    Y0 = S + 1
    S = array(0, c(nt, n, TT))
    for (i in 1:n) for (t in 1:TT) {
        ind = Y0[i, t]
        S[ind, i, t] = 1
    }
    if (is.matrix(X)) 
        X = array(X, c(n, TT, 1))
    nc = dim(X)[3]
    ne = lev - 1
    XX = X
    X = array(0, c(ne, nc, n, TT))
    for (i in 1:n) for (t in 1:TT) {
        X[, , i, t] = rep(1, ne) %o% XX[i, t, ]
    }
    opt = list(TolFun = 10^-6, TolX = 10^-6)
    opt1 = list(TolFun = 10^-6, Display = "iter")
    out = LMest:::marg_param(lev, "g")
    Cm = out$C
    Mm = out$M
    Gm = cbind(-rep(1, lev - 1), diag(lev - 1))
    Hm = rbind(rep(0, lev - 1), diag(lev - 1))
    GHt = t(Gm) %*% t(Hm)
    lm = c(1, rep(0, lev - 1))
    Lm = rbind(rep(0, lev - 1), diag(lev - 1)) - rbind(diag(lev - 1), rep(0, lev - 
        1))
    if (q == 1) 
        sup = 0 else {
        lim = 5
        sup = seq(-lim, lim, 2 * lim/(q - 1))
    }
    Mar = diag(k) %x% matrix(1, 1, q)
    G2 = NULL
    H2 = NULL
    IPI = NULL
    if (k > 1) {
        if (mod == 0 || mod == 2) {
            for (c in 1:k) {
                G2c = diag(k)[, -c]
                H2c = diag(k)[-c, ]
                if (k == 2) 
                  H2c[c] = -1 else H2c[, c] = -1
                if (is.null(G2)) 
                  G2 = G2c else if (k == 2) 
                  G2 = LMest:::blkdiag(matrix(G2, ncol = 1), matrix(G2c, ncol = 1)) else G2 = LMest:::blkdiag(G2, G2c)
                if (is.null(H2)) 
                  H2 = H2c else if (k == 2) 
                  H2 = LMest:::blkdiag(matrix(H2, nrow = 1), matrix(H2c, nrow = 1)) else H2 = LMest:::blkdiag(H2, H2c)
                IPI = c(IPI, c + seq(0, k * (k - 1), k))
            }
        }
        if (mod == 1) {
            G2 = diag(k)[, -1]
            if (k == 2) 
                G2 = matrix(G2, ncol = 1)
            H2 = diag(k)
            H2[, 1] = -1
            H2 = H2[-1, ]
        }
    }
    if (is.null(mu)) {
        Pim = apply(S, c(1, 2), sum) + 0.05 * TT
        Eta = Cm %*% log(Mm %*% Pim)
        Eta = Eta %x% matrix(1, 1, TT)
        eta = as.vector(Eta)
        Z = matrix(aperm(X, c(1, 4, 3, 2)), n * ne * TT, dim(X)[2])
        Z = cbind(matrix(1, n * TT, 1) %x% diag(ne), Z)
        par = ginv(t(Z) %*% Z) %*% t(Z) %*% eta
        mu = par[1:ne]
        par = par[-(1:ne)]
        be = par
        if (k == 1) 
            al = NULL else {
            if (start == 1) 
                al = rnorm(k) * k else al = seq(-k, k, 2 * k/(k - 1))
            mu = mu + al[1]
            al = al[-1] - al[1]
        }
        if (k == 1) 
            PI = 1 else {
            if (mod == 0 || mod == 2) {
                PI = matrix(1, k, k) + 9 * diag(k)
                PI = diag(1/rowSums(PI)) %*% PI
            }
            if (mod == 1) 
                PI = diag(k)
        }
        if (start == 1) {
            la = matrix(runif(k), k, 1)
            la = la/sum(la)
            rho = 2 * matrix(runif(k), k, 1) - 1
            si = runif(1) * 5
        } else {
            la = matrix(1, k, 1)/k
            rho = matrix(0, k, 1)
            si = 3
        }
    }
    par = c(mu, al, si, be)
    if (k == 1) 
        tau = NULL else {
        if (mod == 0 || mod == 2) 
            tau = H2 %*% log(PI[IPI]) else tau = H2 %*% log(la)
    }
    wei = dnorm(sup)
    wei = wei/sum(wei)
    las = as.vector(la %x% wei)
    lrho = (rho + 1)/2
    lrho = log(lrho/(1 - lrho))
    SUP = sup %o% rep(1, q)
    WEI = matrix(1, k * q, k * q)
    if (sup > 0) {
        for (j in 1:k) {
            ind = (j - 1) * q + (1:q)
            Wei = dnorm(t(SUP), rho[j] * SUP, sqrt(1 - rho[j]^2))
            Wei = Wei/rowSums(Wei)
            WEI[, ind] = matrix(1, k, 1) %x% Wei
        }
    }
    PIs = (PI %x% matrix(1, q, q)) * WEI
    t0 = proc.time()
    X1 = matrix(X, ne * nc, n * TT)
    out1 = t(unique(t(X1)))
    nd = ncol(out1)
    indn = rep(0, n * TT)
    INDN = vector("list", nd)
    tmp = ne * nc
    for (jd in 1:nd) {
        ind = which(colSums(X1 == out1[, jd]) == tmp)
        indn[ind] = jd
        INDN[[jd]]$ind = ind
    }
    indn = matrix(indn, n, TT)
    Xd = array(out1, c(ne, nc, nd))
    LLm1 = array(t(Lm), c(ncol(Lm), nrow(Lm), nd))
    itg = 0
    cont = 1
    while (cont && itg < 5) {
        cont = 0
        itg = itg + 1
        I = diag(ne)
        one = matrix(1, ne, 1)
        Pio = array(0, c(n, k * q, TT))
        par0 = par[1:(lev - 1 + k)]
        Eta01 = LMest:::prod_array(Xd, par[(lev + k):length(par)])
        j = 0
        for (c in 1:k) {
            u = matrix(0, 1, k)
            u[c] = 1
            u = u[-1]
            D0 = cbind(I, matrix(u, nrow = 1) %x% one)
            for (d in 1:q) {
                j = j + 1
                D = cbind(D0, sup[d] * one)
                agg = D %*% par0
                Eta1 = Eta01 + agg %*% rep(1, nd)
                Qv1 = expit(Eta1)
                Qv1 = pmin(pmax(Qv1, 10^-100), 1 - 10^-100)
                Pv1 = lm %o% rep(1, nd) + Lm %*% Qv1
                Pv1 = pmin(pmax(Pv1, 10^-100), 1 - 10^-100)
                for (t in 1:TT) Pio[, j, t] = colSums(S[, , t] * Pv1[, indn[, t]])
            }
        }
        Q = rec1(Pio, las, PIs)
        if (q * k == 1) 
            pim = Q[, , TT] else pim = rowSums(Q[, , TT])
        lk = sum(log(pim))
        if (tol > 1) {
            est = NULL
            return
        }
        t0 = proc.time()[1]
        tdisp = 5
        it = 0
        lko = lk - 10^10
        dis = 0
        while ((lk - lko)/abs(lk) > tol & it < maxit) {
            it = it + 1
            lko = lk
            paro = par
            tauo = tau
            lrhoo = lrho
            out = rec3(Q, PIs, Pio, pim)
            U = out$U
            V = out$V
            if (k > 1) {
                if (mod == 0 || mod == 2) {
                  u1 = Mar %*% rowSums(U[, , 1])
                  V1 = Mar %*% V %*% t(Mar)
                  if (mod == 0) {
                    out = optim(tau, lk_sta, gr = NULL, as.vector(u1), V1, G2, outl = FALSE, 
                      method = "BFGS")
                    tau = out$par
                    out = lk_sta(tau, as.vector(u1), V1, G2, outl = TRUE)
                    flk = out$flk
                    la = out$la
                    PI = out$PI
                  } else if (mod == 2) {
                    la = u1/n
                    PI = diag(1/rowSums(V)) %*% V
                  }
                }
                if (mod == 1) {
                  u1 = Mar %*% rowSums(U[, , 1])
                  la = u1/n
                  tau = H2 %*% log(la)
                }
            }
            if (q > 1) {
                for (c in 1:k) {
                  Vc = matrix(0, q, q)
                  ind = (c - 1) * q + (1:q)
                  for (d in 1:k) {
                    ind1 = (d - 1) * q + (1:q)
                    Vc = Vc + V[ind1, ind]
                  }
                  out = optim(lrho[c], lk_ar_rho, gr = NULL, SUP, Vc, outp = FALSE, 
                    method = "BFGS")
                  lrho[c] = out$par
                  out = lk_ar_rho(lrho[c], SUP, Vc, outp = TRUE)
                  flk = out$flk
                  Wei = out$Wei
                  rho[c] = out$rho
                  WEI[, ind] = matrix(1, k, 1) %x% Wei
                }
            }
            las = as.vector(la %x% wei)
            PIs = (PI %x% matrix(1, q, q)) * WEI
            U = aperm(U, c(2, 1, 3))
            s = 0
            FF = 0
            j = 0
            for (c in 1:k) {
                u = matrix(0, 1, k)
                u[c] = 1
                u = u[-1]
                D0 = cbind(I, t(as.matrix(u)) %x% one)
                for (d in 1:q) {
                  j = j + 1
                  D = cbind(D0, sup[d] * one)
                  agg = as.vector(D %*% par0)
                  Eta1 = Eta01 + agg %o% rep(1, nd)
                  Qv1 = expit(Eta1)
                  Qv1 = pmin(pmax(Qv1, 10^-100), 1 - 10^-100)
                  Pit1 = lm %o% rep(1, nd) + Lm %*% Qv1
                  Pit1 = pmin(pmax(Pit1, 10^-100), 1 - 10^-100)
                  QQv1 = Qv1 * (1 - Qv1)
                  DPv1 = 1/Pit1
                  RRtc1 = array(0, c(ne, lev, nd))
                  for (j1 in 1:ne) for (j2 in 1:lev) RRtc1[j1, j2, ] = QQv1[j1, ] * 
                    DPv1[j2, ]
                  RRtc1 = RRtc1 * LLm1
                  XXRi1 = array(0, c(dim(D)[1], dim(D)[2] + dim(Xd)[2], nd))
                  for (h2 in 1:nd) XXRi1[, , h2] = cbind(D, Xd[, , h2])
                  XXRi1 = aperm(XXRi1, c(2, 1, 3))
                  pc = U[, j, ]
                  pc = as.vector(pc)
                  nt = dim(S)[1]
                  YGP = matrix(S, nt, n * TT) - Pit1[, as.vector(indn)]
                  Om = array(0, c(lev, lev, nd))
                  for (r1 in 1:lev) for (r2 in 1:lev) {
                    if (r2 == r1) {
                      Om[r1, r2, ] = Pit1[r1, ] - Pit1[r1, ] * Pit1[r2, ]
                    } else {
                      Om[r1, r2, ] = -Pit1[r1, ] * Pit1[r2, ]
                    }
                  }
                  for (jd in 1:nd) {
                    ind = INDN[[jd]]$ind
                    pci = pc[ind]
                    XRi = (XXRi1[, , jd] %*% RRtc1[, , jd]) %*% GHt
                    if (length(ind) == 1) {
                      s = s + XRi %*% (YGP[, ind] * pci)
                    } else {
                      s = s + XRi %*% (YGP[, ind] %*% pci)
                    }
                    FF = FF + sum(pci) * (XRi %*% Om[, , jd]) %*% t(XRi)
                  }
                }
            }
            dpar = ginv(FF) %*% s
            mdpar = max(abs(dpar))
            if (mdpar > 1) 
                dpar = dpar/mdpar
            par = par + dpar
            si = par[ne + k]
            par0 = par[1:(lev - 1 + k)]
            Eta01 = LMest:::prod_array(Xd, par[(lev + k):length(par)])
            j = 0
            for (c in 1:k) {
                u = matrix(0, 1, k)
                u[c] = 1
                u = u[-1]
                D0 = cbind(I, t(as.matrix(u)) %x% one)
                for (d in 1:q) {
                  j = j + 1
                  D = cbind(D0, sup[d] * one)
                  agg = as.vector(D %*% par0)
                  Eta1 = Eta01 + agg %o% rep(1, nd)
                  Qv1 = expit(Eta1)
                  Qv1 = pmin(pmax(Qv1, 10^-100), 1 - 10^-100)
                  Pv1 = lm %o% rep(1, nd) + Lm %*% Qv1
                  Pv1 = pmin(pmax(Pv1, 10^-100), 1 - 10^-100)
                  for (t in 1:TT) Pio[, j, t] = colSums(S[, , t] * Pv1[, indn[, t]])
                }
            }
            Q = rec1(Pio, las, PIs)
            if (k * q == 1) 
                pim = Q[, , TT] else pim = rowSums(Q[, , TT])
            lk = sum(log(pim))
            dis = max(abs(c(par - paro, tau - tauo, lrho - lrhoo)))
        }
        par1 = NULL
        if (k > 1) 
            par1 = tau
        if (q > 1) 
            par1 = c(par1, lrho)
        par1 = c(par1, par)
        it = 0
        t0 = proc.time()[1]
        while (abs(lk - lko) > 10^-5 && it < 100 && mod == 1) {
            lko = lk
            it = it + 1
            out = lk_obs_manifest(par1, S, Xd, indn, lev, k, sup, G2, IPI, mod, outp = TRUE)
            nx = length(par1)
            d0 = out$s
            ny = length(d0)
            D = matrix(0, nx, ny)
            for (i in 1:nx) {
                o = matrix(0, nx, 1)
                o[i] = 10^-6
                out = lk_obs_manifest(par1 + o, S, Xd, indn, lev, k, sup, G2, IPI, 
                  mod, outp = TRUE)
                d1 = out$s
                d = (d1 - d0)/10^-6
                D[i, ] = t(d)
            }
            s1 = d0
            J1 = D
            J1 = -(J1 + t(J1))/2
            dpar1 = ginv(J1) %*% s1
            mdpar1 = max(abs(dpar1))
            if (mdpar1 > 0.5) 
                dpar1 = dpar1/mdpar1 * 0.5
            par1o = par1
            par1 = par1 + dpar1
            lktmp = lk_obs_manifest(par1, S, Xd, indn, lev, k, sup, G2, IPI, mod)
            lk = lktmp$lk
            cont = 0
            while (lk < (lko - 10^-6)) {
                cont = 1
                dpar1 = dpar1/2
                par1 = par1o + dpar1
                lktmp = lk_obs_manifest(par1, S, Xd, indn, lev, k, sup, G2, IPI, 
                  mod)
                lk = lktmp$lk
            }
            out = trans_par(par1, lev, k, sup, G2, IPI, mod)
            la = out$la
            PI = out$PI
            rho = out$rho
            si = out$si
            par = out$par
            lrho = out$lrho
            tau = out$tau
        }
        if (cont) {
            las = as.vector(la %x% wei)
            PIs = (PI %x% matrix(1, q, q)) * WEI
            WEI = matrix(0, k * q, k * q)
            for (j in 1:k) {
                ind = (j - 1) * q + (1:q)
                Wei = dnorm(t(SUP), rho[j] * SUP, sqrt(1 - rho[j]^2))
                Wei = Wei/rowSums(Wei)
                WEI[, ind] = matrix(1, k, 1) %x% Wei
            }
            PIs = (PI %x% matrix(1, q, q)) * WEI
            tol = tol/2
        }
    }
    mu = par[1:ne]
    al = 0
    if (k > 1) 
        al = c(al, par[(ne + 1):(ne + k - 1)])
    mu = mu + al %*% la
    al = al - al %*% la
    be = par[(ne + k + 1):length(par)]
    if (mod == 0 || mod == 2) 
        np = k * (k - 1)
    if (mod == 1) 
        np = k - 1
    np = np + (ne + (k - 1) + nc) + ((k + 1) * (q > 1))
    if (q == 1) {
        si = NULL
        rho = NULL
    }
    aic = -2 * lk + 2 * (np)
    bic = -2 * lk + log(n) * (np)
    if (output) {
        out = lk_obs_manifest(par1, S, Xd, indn, lev, k, sup, G2, IPI, mod, outp = TRUE)
        lk = out$lk
        U = out$U
        sup1 = t(Mar) %*% al
        if (q > 1) 
            sup1 = sup1 + matrix(1, k, 1) %x% (sup * si)
        PRED0 = array(0, c(n, k, TT))
        PRED1 = matrix(0, n, TT)
        for (t in 1:TT) {
            PRED0[, , t] = U[, , t] %*% t(Mar)
            PRED1[, t] = U[, , t] %*% sup1
        }
    }
    if (out_se) {
        out = lk_obs_manifest(par1, S, Xd, indn, lev, k, sup, G2, IPI, mod, outp = TRUE)
        nx = length(par1)
        d0 = out$s
        ny = length(d0)
        D = matrix(0, nx, ny)
        for (i in 1:nx) {
            o = matrix(0, nx, 1)
            o[i] = 10^-6
            out = lk_obs_manifest(par1 + o, S, Xd, indn, lev, k, sup, G2, IPI, mod, 
                outp = TRUE)
            d1 = out$s
            d = (d1 - d0)/10^-6
            D[i, ] = t(d)
        }
        s1 = d0
        J1 = D
        J1 = -(J1 + t(J1))/2
        se1 = sqrt(diag(ginv(J1)))
        if (k > 1) {
            if (mod == 0 || mod == 2) 
                se1 = se1[-(1:(k * (k - 1)))]
            if (mod == 1) 
                se1 = se1[-(1:(k - 1))]
        }
        if (q == 1) 
            lrho = NULL
        if (q > 1) {
            lrho = se1[1:k]
            se1 = se1[-(1:k)]
        }
        selrho = lrho
        sebe = se1[(ne + k + 1):length(se1)]
    }
    out = list(mu = mu, al = al, be = be, si = si, rho = rho, la = la, PI = PI, lk = lk, 
        np = np, aic = aic, bic = bic, call = match.call())
    if (out_se) {
        out$selrho = selrho
        out$sebe = sebe
        out$J1 = J1
        out$score = s1
    }
    if (output) {
        out$Phi = Pio
        out$PRED0 = PRED0
        out$PRED1 = PRED1
    }
    class(out) = "LMmanifest"
    return(out)
}

# this rewrites function from package LMest so that it is quicker for
#  our purposes
# it is used in lmfirth for computation of score and final estimates 

est_lm_cov_manifest2 = function(S, X, lev, k, q, mod, tol = 10^-8, maxit = 1000, 
    start = 0, mu = NULL, al = NULL, be = NULL, si = NULL, rho = NULL, la = NULL, 
    PI = NULL, output = FALSE, out_se = FALSE) {
    nt = prod(lev)
    n = dim(S)[2]
    TT = dim(S)[3]
    nc = dim(X)[2]
    ne = lev - 1
    opt = list(TolFun = 10^-6, TolX = 10^-6)
    opt1 = list(TolFun = 10^-6, Display = "iter")
    out = LMest:::marg_param(lev, "g")
    Cm = out$C
    Mm = out$M
    Gm = cbind(-rep(1, lev - 1), diag(lev - 1))
    Hm = rbind(rep(0, lev - 1), diag(lev - 1))
    GHt = t(Gm) %*% t(Hm)
    lm = c(1, rep(0, lev - 1))
    Lm = rbind(rep(0, lev - 1), diag(lev - 1)) - rbind(diag(lev - 1), rep(0, lev - 
        1))
    if (q == 1) 
        sup = 0 else {
        lim = 5
        sup = seq(-lim, lim, 2 * lim/(q - 1))
    }
    Mar = diag(k) %x% matrix(1, 1, q)
    G2 = NULL
    H2 = NULL
    IPI = NULL
    if (k > 1) {
        if (mod == 0 || mod == 2) {
            for (c in 1:k) {
                G2c = diag(k)[, -c]
                H2c = diag(k)[-c, ]
                if (k == 2) 
                  H2c[c] = -1 else H2c[, c] = -1
                if (is.null(G2)) 
                  G2 = G2c else if (k == 2) 
                  G2 = LMest:::blkdiag(matrix(G2, ncol = 1), matrix(G2c, ncol = 1)) else G2 = LMest:::blkdiag(G2, G2c)
                if (is.null(H2)) 
                  H2 = H2c else if (k == 2) 
                  H2 = LMest:::blkdiag(matrix(H2, nrow = 1), matrix(H2c, nrow = 1)) else H2 = LMest:::blkdiag(H2, H2c)
                IPI = c(IPI, c + seq(0, k * (k - 1), k))
            }
        }
        if (mod == 1) {
            G2 = diag(k)[, -1]
            if (k == 2) 
                G2 = matrix(G2, ncol = 1)
            H2 = diag(k)
            H2[, 1] = -1
            H2 = H2[-1, ]
        }
    }
    if (is.null(mu)) {
        Pim = apply(S, c(1, 2), sum) + 0.05 * TT
        Eta = Cm %*% log(Mm %*% Pim)
        Eta = Eta %x% matrix(1, 1, TT)
        eta = as.vector(Eta)
        Z = matrix(aperm(X, c(1, 4, 3, 2)), n * ne * TT, dim(X)[2])
        Z = cbind(matrix(1, n * TT, 1) %x% diag(ne), Z)
        par = ginv(t(Z) %*% Z) %*% t(Z) %*% eta
        mu = par[1:ne]
        par = par[-(1:ne)]
        be = par
        if (k == 1) 
            al = NULL else {
            if (start == 1) 
                al = rnorm(k) * k else al = seq(-k, k, 2 * k/(k - 1))
            mu = mu + al[1]
            al = al[-1] - al[1]
        }
        if (k == 1) 
            PI = 1 else {
            if (mod == 0 || mod == 2) {
                PI = matrix(1, k, k) + 9 * diag(k)
                PI = diag(1/rowSums(PI)) %*% PI
            }
            if (mod == 1) 
                PI = diag(k)
        }
        if (start == 1) {
            la = matrix(runif(k), k, 1)
            la = la/sum(la)
            rho = 2 * matrix(runif(k), k, 1) - 1
            si = runif(1) * 5
        } else {
            la = matrix(1, k, 1)/k
            rho = matrix(0, k, 1)
            si = 3
        }
    }
    par = c(mu, al, si, be)
    if (k == 1) 
        tau = NULL else {
        if (mod == 0 || mod == 2) 
            tau = H2 %*% log(PI[IPI]) else tau = H2 %*% log(la)
    }
    wei = dnorm(sup)
    wei = wei/sum(wei)
    las = as.vector(la %x% wei)
    lrho = (rho + 1)/2
    lrho = log(lrho/(1 - lrho))
    SUP = sup %o% rep(1, q)
    WEI = matrix(1, k * q, k * q)
    if (sup > 0) {
        for (j in 1:k) {
            ind = (j - 1) * q + (1:q)
            Wei = dnorm(t(SUP), rho[j] * SUP, sqrt(1 - rho[j]^2))
            Wei = Wei/rowSums(Wei)
            WEI[, ind] = matrix(1, k, 1) %x% Wei
        }
    }
    PIs = (PI %x% matrix(1, q, q)) * WEI
    t0 = proc.time()
    X1 = matrix(X, ne * nc, n * TT)
    out1 = t(unique(t(X1)))
    nd = ncol(out1)
    indn = rep(0, n * TT)
    INDN = vector("list", nd)
    tmp = ne * nc
    for (jd in 1:nd) {
        ind = which(colSums(X1 == out1[, jd]) == tmp)
        indn[ind] = jd
        INDN[[jd]]$ind = ind
    }
    indn = matrix(indn, n, TT)
    Xd = array(out1, c(ne, nc, nd))
    LLm1 = array(t(Lm), c(ncol(Lm), nrow(Lm), nd))
    I = diag(ne)
    one = matrix(1, ne, 1)
    Pio = array(0, c(n, k * q, TT))
    par0 = par[1:(lev - 1 + k)]
    Eta01 = LMest:::prod_array(Xd, par[(lev + k):length(par)])
    j = 0
    for (c in 1:k) {
        u = matrix(0, 1, k)
        u[c] = 1
        u = u[-1]
        D0 = cbind(I, matrix(u, nrow = 1) %x% one)
        for (d in 1:q) {
            j = j + 1
            D = cbind(D0, sup[d] * one)
            agg = D %*% par0
            Eta1 = Eta01 + agg %*% rep(1, nd)
            Qv1 = expit(Eta1)
            Qv1 = pmin(pmax(Qv1, 10^-100), 1 - 10^-100)
            Pv1 = lm %o% rep(1, nd) + Lm %*% Qv1
            Pv1 = pmin(pmax(Pv1, 10^-100), 1 - 10^-100)
            for (t in 1:TT) Pio[, j, t] = colSums(S[, , t] * Pv1[, indn[, t]])
        }
    }
    Q = rec1(Pio, las, PIs)
    if (q * k == 1) 
        pim = Q[, , TT] else pim = rowSums(Q[, , TT])
    lk = sum(log(pim))
    if (tol > 1) {
        est = NULL
        return
    }
    t0 = proc.time()[1]
    tdisp = 5
    it = 0
    lko = lk - 10^10
    dis = 0
    par1 = NULL
    if (k > 1) 
        par1 = tau
    if (q > 1) 
        par1 = c(par1, lrho)
    par1 = c(par1, par)
    if (output) {
        out = lk_obs_manifest(par1, S, Xd, indn, lev, k, sup, G2, IPI, mod, outp = TRUE)
        lk = out$lk
        U = out$U
        sup1 = t(Mar) %*% al
        if (q > 1) 
            sup1 = sup1 + matrix(1, k, 1) %x% (sup * si)
        PRED0 = array(0, c(n, k, TT))
        PRED1 = matrix(0, n, TT)
        for (t in 1:TT) {
            PRED0[, , t] = U[, , t] %*% t(Mar)
            PRED1[, t] = U[, , t] %*% sup1
        }
    }
    if (out_se) {
        out = lk_obs_manifest(par1, S, Xd, indn, lev, k, sup, G2, IPI, mod, outp = TRUE)
        nx = length(par1)
        d0 = out$s
        ny = length(d0)
        func = function(x) {
            lk_obs_manifest(x, S, Xd, indn, lev, k, sup, G2, IPI, mod, outp = TRUE)$s
        }
        D = jacobian(func, par1, method = "simple")
        s1 = d0
        J1 = D
        J1 = -(J1 + t(J1))/2
        ap = apply(J1, 1, sum)
        if (any(ap == 0)) {
            wh = which(apply(J1, 1, sum) == 0)
            J1 = J1[-wh, -wh]
        }
        se1 = sqrt(diag(ginv(J1)))
        if (k > 1) {
            if (mod == 0 || mod == 2) 
                se1 = se1[-(1:(k * (k - 1)))]
            if (mod == 1) 
                se1 = se1[-(1:(k - 1))]
        }
        if (q == 1) 
            lrho = NULL
        if (q > 1) {
            lrho = se1[1:k]
            se1 = se1[-(1:k)]
        }
        selrho = lrho
        sebe = se1[(ne + k + 1):length(se1)]
    }
    out = list(mu = mu, score = s1, al = al, be = be, si = si, rho = rho, la = la, 
        PI = PI, lk = lk, call = match.call())
    if (out_se) {
        out$selrho = selrho
        out$sebe = sebe
        out$J1 = J1
    }
    if (output) {
        out$Phi = Pio
        out$PRED0 = PRED0
        out$PRED1 = PRED1
    }
    class(out) = "LMmanifest"
    return(out)
}

# observed likelihood (simplified and corrected from package LMest) 

stationary = function(tau, k, G2, IPI) {
    Pi = matrix(0, k, k)
    Pi[IPI] = exp(G2 %*% tau)
    Pi = Pi/rowSums(Pi)
    St = Pi
    D0 = array(0, c(k, k, k^2))
    j = 0
    for (c2 in 1:k) for (c1 in 1:k) {
        j = j + 1
        D0[c1, c2, j] = 1
    }
    D = D0
    jnk = 1000
    Stold = Pi[1, ]
    while (jnk > 1e-04) {
        for (j in 1:k^2) {
            D[, , j] = D[, , j] %*% Pi + St %*% D0[, , j]
        }
        Stold = St
        St = St %*% Pi
        jnk = max(abs(St - Stold))
    }
    la = colMeans(St)
    d1 = NULL
    for (j in 1:(k^2)) d1 = rbind(d1, colMeans(D[, , j]))
    Om = diag(Pi[1, ]) - Pi[1, ] %o% Pi[1, ]
    for (c in 2:k) {
        piv = Pi[c, ]
        Om = blkdiag(Om, diag(piv) - piv %o% piv)
    }
    d0 = t(G2) %*% Om
    d1 = d0 %*% d1[IPI, ]
    out = list(d0 = d0, d1 = d1)
    out
}

lk_obs_manifest = function(par, Y, Xd, indn, lev, k, sup, G2, IPI, mod, outp = FALSE) {
    n = dim(Y)[2]
    TT = dim(Y)[3]
    nd = dim(Xd)[3]
    np = lev - 1
    q = length(sup)
    Gm = cbind(-rep(1, lev - 1, 1), diag(lev - 1))
    Hm = rbind(rep(0, lev - 1), diag(lev - 1))
    GHt = t(Gm) %*% t(Hm)
    lm = c(1, rep(0, lev - 1))
    Lm = rbind(rep(0, lev - 1), diag(lev - 1)) - rbind(diag(lev - 1), rep(0, lev - 
        1))
    INDN = vector("list", nd)
    for (jd in 1:nd) INDN[[jd]]$ind = which(as.vector(indn) == jd)
    out = LMest:::trans_par(par, lev, k, sup, G2, IPI, mod)
    la = out$la
    PI = out$PI
    rho = out$rho
    si = out$si
    par = out$par
    lrho = out$lrho
    tau = out$tau
    wei = dnorm(sup)
    wei = wei/sum(wei)
    las = as.vector(la %x% wei)
    if (q > 1) {
        SUP = sup %o% rep(1, q)
        WEI = matrix(0, k * q, k * q)
        for (j in 1:k) {
            ind = (j - 1) * q + (1:q)
            Wei = dnorm(t(SUP), rho[j] * SUP, sqrt(1 - rho[j]^2))
            Wei = Wei/rowSums(Wei)
            WEI[, ind] = matrix(1, k, 1) %x% Wei
        }
        PIs = (PI %x% matrix(1, q, q)) * WEI
    } else PIs = PI
    LLm1 = array(0, c(dim(Lm)[2], dim(Lm)[1], nd))
    for (re in 1:nd) LLm1[, , re] = t(Lm)
    I = diag(np)
    one = matrix(1, np, 1)
    Pio = array(0, c(n, k * q, TT))
    par0 = par[1:(lev - 1 + k)]
    Eta01 = LMest:::prod_array(Xd, par[(lev + k):length(par)])
    j = 0
    for (c in 1:k) {
        u = matrix(0, 1, k)
        u[c] = 1
        u = u[-1]
        D0 = cbind(I, matrix(u, nrow = 1) %x% one)
        for (d in 1:q) {
            j = j + 1
            D = cbind(D0, sup[d] * one)
            agg = D %*% par0
            Eta1 = Eta01 + agg %*% matrix(1, 1, nd)
            Qv1 = expit(Eta1)
            Qv1 = pmin(pmax(Qv1, 10^-100), 1 - 10^-100)
            Pv1 = lm %*% matrix(1, 1, nd) + Lm %*% Qv1
            Pv1 = pmin(pmax(Pv1, 10^-100), 1 - 10^-100)
            for (t in 1:TT) Pio[, j, t] = colSums(Y[, , t] * Pv1[, indn[, t]])
        }
    }
    if (k == 1) 
        PIs = as.matrix(PIs)
    Q = rec1(Pio, las, PIs)
    if (q * k == 1) 
        pim = Q[, , TT] else pim = rowSums(Q[, , TT])
    lk = sum(log(pim))
    if (outp) {
        out = rec3(Q, PIs, Pio, pim)
        U = out$U
        V = out$V
    } else {
        U = NULL
        V = NULL
    }
    s1 = NULL
    if (outp) {
        if (k > 1) {
            if (mod == 0 || mod == 2) {
                out = stationary(tau, k, G2, IPI)
                d21 = out$d0
                d22 = out$d1
                Mar = diag(k) %x% matrix(1, 1, q)
                u1 = Mar %*% rowSums(U[, , 1])
                V1 = Mar %*% V %*% t(Mar)
                s2 = d22 %*% (u1/la) + d21 %*% (V1[IPI]/PI[IPI])
                
            }
            if (mod == 1) {
                Mar = diag(k) %x% matrix(1, 1, q)
                u1 = Mar %*% rowSums(U[, , 1])
                la = as.vector(la)
                Om = diag(la) - la %o% la
                d22 = t(G2) %*% Om
                s2 = d22 %*% (u1/la)
            }
        } else {
            s2 = NULL
        }
        if (q > 1) {
            s3 = matrix(0, k, 1)
            DD1 = matrix(0, k * q, k * q)
            for (j in 1:k) {
                ind = (j - 1) %*% q + (1:q)
                D1 = (rho[j] + SUP * (t(SUP) - rho[j] * SUP))/(1 - rho[j]^2) - (t(SUP) - 
                  rho[j] * SUP)^2 * rho[j]/(1 - rho[j]^2)^2
                Wei = dnorm(t(SUP), rho[j] * SUP, sqrt(1 - rho[j]^2))
                D1 = D1 - (rowSums(Wei * D1)/rowSums(Wei)) %*% matrix(1, 1, q)
                DD1[, ind] = matrix(1, k, 1) %x% D1
                for (j1 in 1:k) {
                  ind1 = (j1 - 1) %*% q + (1:q)
                  s3[j] = s3[j] + 2 * exp(lrho[j])/(1 + exp(lrho[j]))^2 * sum(sum(V[ind1, 
                    ind] * D1))
                }
            }
        } else {
            s3 = NULL
        }
        U = aperm(U, c(2, 1, 3))
        s4 = 0
        j = 0
        for (c in 1:k) {
            u = matrix(0, 1, k)
            u[c] = 1
            u = u[-1]
            D0 = cbind(I, matrix(u, nrow = 1) %x% one)
            for (d in 1:q) {
                j = j + 1
                D = cbind(D0, sup[d] * one)
                agg = D %*% par0
                Eta1 = Eta01 + agg %*% matrix(1, 1, nd)
                Qv1 = expit(Eta1)
                Qv1 = pmin(pmax(Qv1, 10^-100), 1 - 10^-100)
                Pit1 = lm %*% matrix(1, 1, nd) + Lm %*% Qv1
                Pit1 = pmin(pmax(Pit1, 10^-100), 1 - 10^-100)
                QQv1 = Qv1 * (1 - Qv1)
                DPv1 = 1/Pit1
                RRtc1 = array(0, c(np, lev, nd))
                for (j1 in 1:np) for (j2 in 1:lev) RRtc1[j1, j2, ] = QQv1[j1, ] * 
                  DPv1[j2, ]
                RRtc1 = RRtc1 * LLm1
                XXRi1 = array(0, c(dim(D)[1], dim(D)[2] + dim(Xd)[2], nd))
                for (h2 in 1:nd) XXRi1[, , h2] = cbind(D, Xd[, , h2])
                XXRi1 = aperm(XXRi1, c(2, 1, 3))
                pc = U[, j, ]
                pc = as.vector(pc)
                nt = dim(Y)[1]
                YGP = matrix(Y, nt, n * TT) - Pit1[, as.vector(indn)]
                for (jd in 1:nd) {
                  ind = INDN[[jd]]$ind
                  pci = pc[ind]
                  XRi = (XXRi1[, , jd] %*% RRtc1[, , jd]) %*% GHt
                  if (length(ind) == 1) {
                    s4 = s4 + XRi %*% (YGP[, ind] * pci)
                  } else {
                    s4 = s4 + XRi %*% (YGP[, ind] %*% pci)
                  }
                }
            }
        }
        s = c(s1, s2, s3, s4)
    } else {
        s = NULL
    }
    out = list(lk = lk, U = U, s = s)
    out
} 

  invglob=function (eta) 
{
    b = length(eta)
    if (b == 1) 
        p = c(1, exp(eta))/(1 + exp(eta))
    else {
        pc = c(1/(1 + exp(eta)), 1)
        Di = cbind(rep(0, b), diag(b)) - cbind(diag(b), rep(0, 
            b))
        Di = rbind(c(1, rep(0, b)), Di)
        p = Di %*% pc
    }
p}


gendata=function(n, TT, k, mu=NULL, xi=NULL, PI=NULL, pi=NULL) {
    lev = 3 
    p=2 
    S=matrix(NA,n,TT) 
    X=array(NA,dim=c(n,TT,p))
    if(is.null(mu)) {mu=seq(-5,5,length=lev-1)}
    if(is.null(xi)) {xi=seq(-5,5,length=k)}
    if(is.null(PI)) {
        PI=diag(k)
        PI=(PI+0.2)/apply(PI+0.2,1,sum)}
    if(is.null(pi)) {pi=rep(1/k,k)}


    X[,,1]=rbinom(n,1,0.5)
    X[,,2]=rnorm(n)

    beta=rep(1,dim(X)[3])

    u=sample(k,n,replace=TRUE,prob=pi)

    glogit=cbind(mu[2]+xi[u]+X[,1,]%*%beta,mu[1]+xi[u]+X[,1,]%*%beta)
    p=apply(glogit,1,invglob)

    for(i in 1:n) {S[i,1]=sample(lev[1],1,prob=p[,i])}

    for(ti in 2:TT) {
        linpred=X[,ti,]%*%beta
        for(i in 1:n) {
            u[i]=sample(k,1,prob=PI[u[i],])
            p=invglob(c(mu[2]+xi[u[i]]+linpred[i],mu[1]+xi[u[i]]+linpred[i]))
            S[i,ti]=sample(lev,1,prob=p)}}
    S=S-1
return(list(S=S,X=X))}

    
